#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <omp.h>
#include "param.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    /* Some initialization */
    int i, j, k, l;
    double *x = mxGetPr(prhs[0]);
    int n = mxGetN(prhs[0]);
    int d = mxGetM(prhs[0]);
    double *y = mxGetPr(prhs[1]);
    double *sig = mxGetPr(prhs[2]);

    int ds = (d*(d+1))/2;
    double *SIG = (double*)calloc(ds,sizeof(double));
    for(l=0;l<d;l++){ /* Precompute the exponential for speed up*/
        SIG[IND2VAR(l,l,d)] = exp(sig[IND2VAR(l,l,d)]);
        for(k=l+1;k<d;k++)
            SIG[IND2VAR(l,k,d)] = sig[IND2VAR(l,k,d)];
    } 

    double *J, *dJ;
    plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL);
    plhs[1]=mxCreateDoubleMatrix(ds, 1, mxREAL);
    J =  mxGetPr(plhs[0]);
    dJ =  mxGetPr(plhs[1]);
    
    int Ki, tm;
    double sum=0, sum1=0;
    double Ktp,temp;
    double Kin=0;
    double reduc0, reduc1;
    
    double **sum2 = (double **)calloc(MAX_THREADS, sizeof(double*));
    double **sum3 = (double **)calloc(MAX_THREADS, sizeof(double*));
    for(i=0;i<MAX_THREADS;i++){
        sum2[i]=(double*)calloc(ds, sizeof(double));
        sum3[i]=(double*)calloc(ds, sizeof(double ));
        for(j=0;j<ds;j++){
            sum2[i][j]=0.0;
            sum3[i][j]=0.0;
        }
    }
    
    /* Compute the objective and gradient function */
    #pragma omp parallel for private(i, j, k, l, Ki, Ktp, tm, temp) shared(x, y, sum2, sum3,d,n) reduction(+:sum, sum1,Kin) num_threads(MAX_THREADS) schedule(dynamic)
    for(i=0;i<n;i++){
        tm=omp_get_thread_num();/* Get the thread number*/ 
        sum +=1;
        sum1 +=1;
        Kin +=1;
        for(j=i+1;j<n;j++){
            Ki = (y[i] == y[j]) ? 1 : 0;
            Kin += 2*Ki;
            Ktp = 0;
            /* Compute the kernel */
            for(l=0;l<d;l++){
                temp = (x[i*d+l]-x[j*d+l])*SIG[IND2VAR(l, l, d)];
                for(k=l+1;k<d;k++){
                    temp += (x[i*d+k]-x[j*d+k])*SIG[IND2VAR(l, k, d)];
                }
                Ktp += (temp*temp);
            }
            Ktp = exp(-0.5*Ktp);
            
            sum += 2*Ktp*Ki;
            sum1 += 2*Ktp*Ktp;
            
            for(l=0;l<d;l++){/*Gradients*/
                temp = SIG[IND2VAR(l, l, d)]*(x[i*d+l]-x[j*d+l]);
                for(k=l+1;k<d;k++)
                    temp += SIG[IND2VAR(l, k, d)]*(x[i*d+k]-x[j*d+k]);
                /* l==k*/
                sum2[tm][IND2VAR(l, l, d)] += 2*SIG[IND2VAR(l, l, d)]*(x[i*d+l]-x[j*d+l])*temp*Ktp*Ki;
                sum3[tm][IND2VAR(l, l, d)] += 2*SIG[IND2VAR(l, l, d)]*(x[i*d+l]-x[j*d+l])*temp*Ktp*Ktp;
                /* l<k<d */
                for(k=l+1;k<d;k++){
                    sum2[tm][IND2VAR(l, k, d)] += 2*Ktp*Ki*(x[i*d+k]-x[j*d+k])*temp;
                    sum3[tm][IND2VAR(l, k, d)] += 2*Ktp*Ktp*(x[i*d+k]-x[j*d+k])*temp;
                }
            }
        }
    }
    J[0] = 1-sum/sqrt(sum1)/sqrt(Kin);
    
    for(l=0;l<d;l++){
        for(k=l;k<d;k++){
            reduc0=0;
            reduc1=0;
            for(i=0;i<MAX_THREADS;i++){
                reduc0+=sum2[i][IND2VAR(l, k, d)];
                reduc1+=sum3[i][IND2VAR(l, k, d)];
            }
            dJ[IND2VAR(l, k, d)] = (reduc0/sqrt(sum1)-sum*reduc1/sqrt(sum1)/sum1)/sqrt(Kin);
        }
    }
    
    /* Free the memory */
    for(i=0;i<MAX_THREADS;i++){
        free(sum2[i]);
        free(sum3[i]);
    }
    free(sum2);
    free(sum3);
    free(SIG);
}