#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <omp.h>
#include "param.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    /* Some initialization */
    int i, j, k, l;
    double *x = mxGetPr(prhs[0]);
    int n = mxGetN(prhs[0]);
    int d = mxGetM(prhs[0]);
    double *y = mxGetPr(prhs[1]);
    double *sig = mxGetPr(prhs[2]);
    
    int ds = (d*(d+1))/2;
    double *SIG = (double*)calloc(ds, sizeof(double));
    for(l=0;l<d;l++){ /* Precompute the exponential for speed up*/
        SIG[IND2VAR(l, l, d)] = exp(sig[IND2VAR(l, l, d)]);
        for(k=l+1;k<d;k++)
            SIG[IND2VAR(l, k, d)] = sig[IND2VAR(l, k, d)];
    }

    double *J, *dJ;
    plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL);
    plhs[1]=mxCreateDoubleMatrix(ds, 1, mxREAL);
    J =  mxGetPr(plhs[0]);
    dJ =  mxGetPr(plhs[1]);
    
    int Ki, tm;
    double sum=0, sum1=0, sum2=0, sum3=0, sum4=0;
    double Ktp, temp;
    double Kin=0;
    double Ktemp;
    double *dKb = (double*)calloc(ds, sizeof(double));
    double **reduc_sum3 = (double**)calloc(MAX_THREADS, sizeof(double*));
    double **reduc_sum4 = (double**)calloc(MAX_THREADS, sizeof(double*));
    double **reduc_dKb = (double**)calloc(MAX_THREADS, sizeof(double*));
    for(i=0;i<MAX_THREADS;i++){
        reduc_dKb[i]=(double*)calloc(ds, sizeof(double));
        reduc_sum3[i]=(double*)calloc(ds, sizeof(double));
        reduc_sum4[i]=(double*)calloc(ds, sizeof(double));
        for(j=0;j<ds;j++){
            reduc_dKb[i][j]=0.0;
            reduc_sum3[i][j]=0.0;
            reduc_sum4[i][j]=0.0;
        }
    }
    
    /* Compute the mean K, dK and K_i */
    double Kb=0, Kib=0;
    #pragma omp parallel for private(i, j, k, l, Ki, Ktp, tm, temp) shared(x, y, reduc_dKb) reduction(+:Kb, Kib) num_threads(MAX_THREADS) schedule(runtime)
    for(i=0;i<n;i++){
        tm = omp_get_thread_num();
        Kb ++;
        Kib ++;
        for(j=i+1;j<n;j++){
            Ki = (y[i] == y[j]) ? 1 : 0;
            Ktp = 0;
            /* Compute the kernel */
            for(l=0;l<d;l++){
                temp = (x[i*d+l]-x[j*d+l])*SIG[IND2VAR(l, l, d)];
                for(k=l+1;k<d;k++){
                    temp += (x[i*d+k]-x[j*d+k])*SIG[IND2VAR(l, k, d)];
                }
                Ktp += (temp*temp);
            }
            Ktp = exp(-0.5*Ktp);
            
            Kb += 2*Ktp;
            Kib += 2*Ki;
            for(l=0;l<d;l++){
                temp =  SIG[IND2VAR(l, l, d)]*(x[i*d+l]-x[j*d+l]);
                for(k=l+1;k<d;k++)
                    temp += SIG[IND2VAR(l, k, d)]*(x[i*d+k]-x[j*d+k]);
                /*l==k*/
                reduc_dKb[tm][IND2VAR(l, l, d)] += 2*SIG[IND2VAR(l, l, d)]*(x[i*d+l]-x[j*d+l])*temp*Ktp;
                /*l<k<d*/
                for(k=l+1;k<d;k++)
                    reduc_dKb[tm][IND2VAR(l, k, d)] += 2*(x[i*d+k]-x[j*d+k])*temp*Ktp;
            }
        }
    }
    Kb = Kb/n/n;
    Kib = Kib/n/n;
    for(k=0;k<ds;k++){
        for(i=0;i<MAX_THREADS;i++)
            dKb[k] += reduc_dKb[i][k];
        dKb[k] = -dKb[k]/n/n;
    }
    /* Free the memory */
    for(i=0;i<MAX_THREADS;i++)
        free(reduc_dKb[i]);
    free(reduc_dKb);
    
    /* Compute the objective and gradient function */
    #pragma omp parallel for private(i, j, k, l, Ki, Ktp, tm, temp) shared(x, y, reduc_sum3, reduc_sum4) reduction(+:sum, sum1, sum2) num_threads(MAX_THREADS) schedule(runtime)
    for(i=0;i<n;i++){
        tm = omp_get_thread_num();
        sum += (1-Kb)*(1-Kb);
        sum1 += (1-Kib)*(1-Kib);
        sum2 += (1-Kib)*(1-Kb);
        
        for(l=0;l<ds;l++){
            reduc_sum3[tm][l] += -(1-Kib)*dKb[l];
            reduc_sum4[tm][l] += -(1-Kb)*dKb[l];
        }
        for(j=i+1;j<n;j++){
            Ki = (y[i] == y[j]) ? 1 : 0;
            Ktp = 0;
            /* Compute the kernel */
            for(l=0;l<d;l++){
                temp = (x[i*d+l]-x[j*d+l])*SIG[IND2VAR(l, l, d)];
                for(k=l+1;k<d;k++){
                    temp += (x[i*d+k]-x[j*d+k])*SIG[IND2VAR(l, k, d)];
                }
                Ktp += (temp*temp);
            }
            Ktp = exp(-0.5*Ktp);
            
            sum += 2*(Ktp-Kb)*(Ktp-Kb);
            sum1 += 2*(Ki-Kib)*(Ki-Kib);
            sum2 += 2*(Ki-Kib)*(Ktp-Kb);
            
            for(l=0;l<d;l++){
                temp =  SIG[IND2VAR(l, l, d)]*(x[i*d+l]-x[j*d+l]);
                for(k=l+1;k<d;k++)
                    temp += SIG[IND2VAR(l, k, d)]*(x[i*d+k]-x[j*d+k]);
                /* l==k*/
                reduc_sum3[tm][IND2VAR(l, l, d)] += 2*(Ki-Kib)*(SIG[IND2VAR(l, l, d)]*(x[i*d+l]-x[j*d+l])*temp*Ktp +dKb[IND2VAR(l, l, d)]);
                reduc_sum4[tm][IND2VAR(l, l, d)] += 2*(Ktp-Kb)*(SIG[IND2VAR(l, l, d)]*(x[i*d+l]-x[j*d+l])*temp*Ktp +dKb[IND2VAR(l, l, d)]);
                for(k=l+1;k<d;k++){
                    reduc_sum3[tm][IND2VAR(l, k, d)] += 2*(Ki-Kib)*((x[i*d+k]-x[j*d+k])*temp*Ktp +dKb[IND2VAR(l, k, d)]);
                    reduc_sum4[tm][IND2VAR(l, k, d)] += 2*(Ktp-Kb)*((x[i*d+k]-x[j*d+k])*temp*Ktp +dKb[IND2VAR(l, k, d)]);
                }
            }
        }
    }
    J[0]=1-sum2/sqrt(sum)/sqrt(sum1);
    
    for(k=0;k<ds;k++){
        sum3=0;sum4=0;
        for(i=0;i<MAX_THREADS;i++){
            sum3+=reduc_sum3[i][k];
            sum4+=reduc_sum4[i][k];
        }
        dJ[k]= (sum3/sqrt(sum)-sum2*sum4/sqrt(sum)/sum)/sqrt(sum1);
    }
    
    /* Free the memory */
    for(i=0;i<MAX_THREADS;i++){
        free(reduc_sum3[i]);
        free(reduc_sum4[i]);
    }
    free(reduc_sum3);
    free(reduc_sum4);
    free(SIG);
}