#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <omp.h>
#include "param.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    /* Some initialization */
    int i, j, k,l;
    double *x = mxGetPr(prhs[0]);
    int n = mxGetN(prhs[0]);
    int d = mxGetM(prhs[0]);
    double *y = mxGetPr(prhs[1]);
    double *sig = mxGetPr(prhs[2]);
    
    int ds = (d*(d+1))/2;
    double *SIG = (double*)calloc(ds,sizeof(double));
    for(l=0;l<d;l++){ /* Precompute the exponential for speed up*/
        SIG[IND2VAR(l,l,d)] = exp(sig[IND2VAR(l,l,d)]);
        for(k=l+1;k<d;k++)
            SIG[IND2VAR(l,k,d)] = sig[IND2VAR(l,k,d)];
    }   
        
    double *J, *dJ;
    plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL);
    plhs[1]=mxCreateDoubleMatrix(ds, 1, mxREAL);
    J =  mxGetPr(plhs[0]);
    dJ =  mxGetPr(plhs[1]);
    
    int Ki, tm;
    double sum=0;
    double Ktp, temp;
    double ** sum1 = (double**)calloc(MAX_THREADS, sizeof(double*));
    for(i=0;i<MAX_THREADS;i++){
        sum1[i]=(double*)calloc(ds, sizeof(double));
        for(j=0;j<ds;j++)
            sum1[i][j]=0.0;
        }
    
    /* Compute the objective and gradient function */
    #pragma omp parallel for private(i, j, k, l, Ki, Ktp, tm, temp) shared(x, y, sum1) reduction(+:sum) num_threads(MAX_THREADS) schedule(runtime)
    for(i=0;i<n;i++){
        for(j=i+1;j<n;j++){
            tm=omp_get_thread_num();/* Get the thread number*/
            Ki = (y[i] == y[j]) ? 1 : 0;
            Ktp = 0;
            /* Compute the kernel */
            for(l=0;l<d;l++){
                temp = (x[i*d+l]-x[j*d+l])*SIG[IND2VAR(l,l,d)];
                for(k=l+1;k<d;k++){
                    temp += (x[i*d+k]-x[j*d+k])*SIG[IND2VAR(l,k,d)];
                }
                Ktp += (temp*temp);
            }
            Ktp = exp(-0.5*Ktp);
    
            sum += 2*(Ktp-Ki)*(Ktp-Ki); /* Objective */
            
            for(l=0;l<d;l++){/* Gradient */
                temp=exp(sig[IND2VAR(l,l,d)])*(x[i*d+l]-x[j*d+l]);
                for(k=l+1;k<d;k++)
                    temp += sig[IND2VAR(l,k,d)]*(x[i*d+k]-x[j*d+k]);
                /* l==k*/
                sum1[tm][IND2VAR(l,l,d)] += 2*SIG[IND2VAR(l,l,d)]*(x[i*d+l]-x[j*d+l])*temp*Ktp*(Ktp-Ki);
                /* l<k<d */
                for(k=l+1;k<d;k++)
                    sum1[tm][IND2VAR(l,k,d)] += 2*(x[i*d+k]-x[j*d+k])*temp*Ktp*(Ktp-Ki);
                
            }
        }
    }
    J[0] = sum/n/n;

    /* Do the reduction manually*/
    for(l=0;l<d;l++){
        for(k=l;k<d;k++){
            sum=0.0;
            for(i=0;i<MAX_THREADS;i++)
                sum += sum1[i][IND2VAR(l, k, d)];
            dJ[IND2VAR(l, k, d)] = -2*sum/n/n;
        }
    }
    
    /*Free the memroy*/
    for(i=0;i<MAX_THREADS;i++)
        free(sum1[i]);
    free(sum1);
    free(SIG);
}