#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <omp.h>
#include "param.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    /* Some initialization */
    int i, j, k;
    double *x = mxGetPr(prhs[0]);
    int n = mxGetN(prhs[0]);
    int d = mxGetM(prhs[0]);
    double *y = mxGetPr(prhs[1]);
    double *sig = mxGetPr(prhs[2]);
    
    double *J, *dJ;
    plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL);
    plhs[1]=mxCreateDoubleMatrix(d, 1, mxREAL);
    J =  mxGetPr(plhs[0]);
    dJ =  mxGetPr(plhs[1]);
    
    int Ki,tm;
    double sum=0, sum1=0;
    double Ktp;
    double Kin=0;
    double reduc0,reduc1;
    
    double **sum2 = (double **)calloc(MAX_THREADS, sizeof(double*));
    double **sum3 = (double **)calloc(MAX_THREADS, sizeof(double*));
    for(i=0;i<MAX_THREADS;i++){
        sum2[i]=(double*)calloc(d,sizeof(double));
        sum3[i]=(double*)calloc(d,sizeof(double ));
    }

    double *EXPSIG = (double*)calloc(d,sizeof(double));
    for(k=0;k<d;k++)/* Pre compute exp(sig) for speed up*/
        EXPSIG[k]=exp(sig[k]);
            
    /* Compute the objective and gradient function */
    #pragma omp parallel for private(i,j,k,Ki,Ktp,tm) shared(x,y,sum2,sum3) reduction(+:sum,sum1,Kin) num_threads(MAX_THREADS) schedule(runtime)   
    for(i=0;i<n;i++){
        sum +=1;
        sum1 +=1;
        Kin +=1;
        for(j=i+1;j<n;j++){
            tm=omp_get_thread_num();/* Get the thread number*/
            Ki = (y[i] == y[j]) ? 1 : 0;
            Kin += 2*Ki;
            Ktp = 0;
            for(k=0;k<d;k++)
                Ktp += (x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k])*EXPSIG[k];
            Ktp = exp(-0.5*Ktp);
            
            sum += 2*Ktp*Ki;
            sum1 += 2*Ktp*Ktp;
            
            for(k=0;k<d;k++){
                sum2[tm][k] += 2*Ktp*Ki*(x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k]);
                sum3[tm][k] += 2*Ktp*Ktp*(x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k]);
            }
        }
    }
    J[0] = 1 -sum/sqrt(sum1)/sqrt(Kin);
    
    for(k=0;k<d;k++){
        reduc0=0;
        reduc1=0;
        for(i=0;i<MAX_THREADS;i++){
            reduc0+=sum2[i][k];
            reduc1+=sum3[i][k];
        }
        dJ[k] = (reduc0/sqrt(sum1)-sum*reduc1/sqrt(sum1)/sum1)*EXPSIG[k]/sqrt(Kin)/2;
    }
    /* Free the memory */
    for(i=1;i<MAX_THREADS;i++){
        free(sum2[i]);
        free(sum3[i]);
    }
    free(sum2);
    free(sum3);
    free(EXPSIG);
}