#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <omp.h>
#include "param.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    /* Some initialization */
    int i, j, k;
    double *x = mxGetPr(prhs[0]);
    int n = mxGetN(prhs[0]);
    int d = mxGetM(prhs[0]);
    double *y = mxGetPr(prhs[1]);
    double *sig = mxGetPr(prhs[2]);

    double *J, *dJ;
    plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL);
    plhs[1]=mxCreateDoubleMatrix(d, 1, mxREAL);
    J =  mxGetPr(plhs[0]);
    dJ =  mxGetPr(plhs[1]);
    
    int Ki, tm;
    double sum=0, sum1=0, sum2=0,sum3=0,sum4=0;
    double Ktp;
    double Kin=0;
    double Ktemp;
    double *dKb = (double*)calloc(d, sizeof(double));
    
    double **reduc_dKb = (double**)calloc(MAX_THREADS, sizeof(double*));
    double **reduc_sum3 = (double**)calloc(MAX_THREADS, sizeof(double*));
    double **reduc_sum4 = (double**)calloc(MAX_THREADS, sizeof(double*));
    for(i=0;i<MAX_THREADS;i++){
        reduc_dKb[i]=(double*)calloc(d, sizeof(double));
        reduc_sum3[i]=(double*)calloc(d, sizeof(double));
        reduc_sum4[i]=(double*)calloc(d, sizeof(double));
    }
    double *EXPSIG = (double*)calloc(d,sizeof(double));
    for(k=0;k<d;k++)/* Pre compute exp(sig) for speed up*/
        EXPSIG[k]=exp(sig[k]);
    
    /* Compute the mean K, dK and K_i */
    double Kb=0, Kib=0;
    #pragma omp parallel for private(i, j, k, Ki, Ktp,tm) shared(x, y, reduc_dKb) reduction(+:Kb, Kib) num_threads(MAX_THREADS) schedule(runtime)
    for(i=0;i<n;i++){
        tm = omp_get_thread_num();
        Kb ++;
        Kib ++;
        for(j=i+1;j<n;j++){
            Ki = (y[i] == y[j]) ? 1 : 0;
            Ktp = 0;
            for(k=0;k<d;k++)
                Ktp += (x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k])*EXPSIG[k];
            Ktp =exp(-0.5*Ktp);
            
            Kb += 2*Ktp;
            Kib += 2*Ki;
            for(k=0;k<d;k++)
                reduc_dKb[tm][k] += 2*(x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k])*Ktp;
        }
    }
    Kb = Kb/n/n;
    Kib = Kib/n/n;
    for(k=0;k<d;k++){
        for(i=0;i<MAX_THREADS;i++)
            dKb[k] += reduc_dKb[i][k];
        dKb[k] = -dKb[k]*EXPSIG[k]/2/n/n;
    }
    /* Free the memory */
    for(i=0;i<MAX_THREADS;i++)
        free(reduc_dKb[i]);
    free(reduc_dKb);
    
    /* Compute the objective and gradient function */
    #pragma omp parallel for private(i, j, k, Ki, Ktp, tm) shared(x, y, reduc_sum3, reduc_sum4) reduction(+:sum, sum1, sum2) num_threads(MAX_THREADS) schedule(runtime)
    for(i=0;i<n;i++){
        tm = omp_get_thread_num();
        sum += (1-Kb)*(1-Kb);
        sum1 += (1-Kib)*(1-Kib);
        sum2 += (1-Kib)*(1-Kb);
        
        for(k=0;k<d;k++){
            reduc_sum3[tm][k] += 2*(1-Kib)*dKb[k];
            reduc_sum4[tm][k] += 2*(1-Kb)*dKb[k];
        }
        
        for(j=i+1;j<n;j++){
            Ki = (y[i] == y[j]) ? 1 : 0;
            Ktp = 0;
            for(k=0;k<d;k++)
                Ktp += (x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k])*EXPSIG[k];
            Ktp = exp(-0.5*Ktp);
            
            sum += 2*(Ktp-Kb)*(Ktp-Kb);
            sum1 += 2*(Ki-Kib)*(Ki-Kib);
            sum2 += 2*(Ki-Kib)*(Ktp-Kb);
            
            for(k=0;k<d;k++){
                reduc_sum3[tm][k] += 2*(Ki-Kib)*((x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k])*EXPSIG[k]*Ktp +2*dKb[k]);
                reduc_sum4[tm][k] += 2*(Ktp-Kb)*((x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k])*EXPSIG[k]*Ktp +2*dKb[k]);
            }
        }
    }

    J[0] = 1-sum2/sqrt(sum)/sqrt(sum1);
    for(k=0;k<d;k++){
        sum3=0;sum4=0;
        for(i=0;i<MAX_THREADS;i++){
            sum3+=reduc_sum3[i][k];
            sum4+=reduc_sum4[i][k];
        }
        dJ[k]= (sum3/sqrt(sum) -sum2*sum4/sqrt(sum)/sum)/sqrt(sum1)/2;
    }
    /* Free the memory */
    for(i=0;i<MAX_THREADS;i++){
        free(reduc_sum3[i]);
        free(reduc_sum4[i]);
    }
    free(reduc_sum3);
    free(reduc_sum4);
    free(EXPSIG);
}