#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <omp.h>
#include "param.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    /* Some initialization */
    int i, j, k;
    double *x = mxGetPr(prhs[0]);
    int n = mxGetN(prhs[0]);
    int d = mxGetM(prhs[0]);
    double *y = mxGetPr(prhs[1]);
    double *sig = mxGetPr(prhs[2]);
    
    double *J, *dJ;
    plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL);
    plhs[1]=mxCreateDoubleMatrix(d, 1, mxREAL);
    J =  mxGetPr(plhs[0]);
    dJ =  mxGetPr(plhs[1]);
    
    int Ki, tm;
    double sum=0;
    double Ktp;
    double ** sum1 = (double**)calloc(MAX_THREADS, sizeof(double*));
    for(i=0;i<MAX_THREADS;i++)
        sum1[i]=(double*)calloc(d, sizeof(double));

    double *EXPSIG = (double*)calloc(d,sizeof(double));
    for(k=0;k<d;k++)/* Pre compute exp(sig) for speed up*/
        EXPSIG[k]=exp(sig[k]);
    
    /* Compute the objective and gradient function */
    #pragma omp parallel for private(i, j, k, Ki, Ktp, tm) shared(x, y, sum1) reduction(+:sum) num_threads(MAX_THREADS) schedule(runtime)
    for(i=0;i<n;i++){
        for(j=i+1;j<n;j++){
            tm=omp_get_thread_num();/* Get the thread number*/
            Ki = (y[i] == y[j]) ? 1 : 0;
            Ktp = 0;
            for(k=0;k<d;k++)
                Ktp += (x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k])*EXPSIG[k];
            Ktp = exp(-0.5*Ktp);;
            
            sum += 2*(Ktp-Ki)*(Ktp-Ki); /* Objective */
            
            for(k=0;k<d;k++)
                sum1[tm][k] += 2*Ktp*(Ktp-Ki)*(x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k]);/* Gradient */
        }
    }
    J[0] = sum/n/n;
    
    /* Do the reduction manually*/
    for(k=0;k<d;k++){
        sum=0;
        for(i=0;i<MAX_THREADS;i++)
            sum += sum1[i][k];
        dJ[k] = -sum*EXPSIG[k]/n/n; /* do the loop */
    }
    for(i=0;i<MAX_THREADS;i++)
        free(sum1[i]);
    free(sum1);
    free(EXPSIG);
}