#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <omp.h>
#include "param.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    /* Some initialization */
    int i, j, k;
    double *x = mxGetPr(prhs[0]);
    int n = mxGetN(prhs[0]);
    int d = mxGetM(prhs[0]);
    double *y = mxGetPr(prhs[1]);
    double *sig = mxGetPr(prhs[2]);
    double EXPSIG = exp(sig[0]); /* Pre compute exp(sig) for speeding up*/
    
    double *J, *dJ;
    plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL);
    plhs[1]=mxCreateDoubleMatrix(1, 1, mxREAL);
    J =  mxGetPr(plhs[0]);
    dJ =  mxGetPr(plhs[1]);
    
    int Ki;
    double sum=0, sum1=0, sum2=0, sum3=0;
    double Ktp, logKtp;
    double Kin=0;
    
    /* Compute the objective and gradient function */
    #pragma omp parallel for private(i, j, k, Ki, Ktp, logKtp) shared(x, y) reduction(+:sum, sum1, sum2, sum3, Kin) num_threads(MAX_THREADS) schedule(runtime)
    for(i=0;i<n;i++){
        sum1 +=1;
        sum3 +=1;
        Kin +=1;
        for(j=i+1;j<n;j++){
            Ki = (y[i] == y[j]) ? 1 : 0;
            Kin += 2*Ki;
            Ktp = 0;
            for(k=0;k<d;k++)
                Ktp += (x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k]);
            logKtp = -0.5*Ktp*EXPSIG;
            Ktp = exp(logKtp);
            
            sum  += 2*(Ktp*logKtp*Ki);
            sum1 += 2*(Ktp*Ktp);
            sum2 += 2*(logKtp*Ktp*Ktp);
            sum3 += 2*(Ktp*Ki);
        }
    }
    J[0] = 1-sum3/sqrt(sum1)/sqrt(Kin);
    dJ[0] =-(sum/sqrt(sum1)-sum2*sum3/sqrt(sum1)/sum1)/sqrt(Kin);
}