#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <omp.h>
#include "param.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    /* Some initialization */
    int i, j, k;
    double *x = mxGetPr(prhs[0]);
    int n = mxGetN(prhs[0]);
    int d = mxGetM(prhs[0]);
    double *y = mxGetPr(prhs[1]);
    double *sig = mxGetPr(prhs[2]);
    double EXPSIG = exp(sig[0]); /* Pre compute exp(sig) for speeding up*/

    double *J, *dJ;
    plhs[0]=mxCreateDoubleMatrix(1, 1, mxREAL);
    plhs[1]=mxCreateDoubleMatrix(1, 1, mxREAL);
    J =  mxGetPr(plhs[0]);
    dJ =  mxGetPr(plhs[1]);
    
    int Ki;
    double sum=0, sum1=0, sum2=0, sum3=0,sum4=0;
    double Ktp, logKtp;
    double Kin=0;
    double Ktemp;
    
    /* Compute the mean K, dK and K_i */
    double Kb=0, dKb=0,Kib=0;
    #pragma omp parallel for private(i,j,k,Ki,Ktp,Ktemp) shared(x,y) reduction(+:Kb,dKb,Kib) num_threads(MAX_THREADS) schedule(runtime)  
    for(i=0;i<n;i++){
        Kb ++;
        Kib ++;
        for(j=i+1;j<n;j++){
            Ki = (y[i] == y[j]) ? 1 : 0;
            Ktp = 0;
            for(k=0;k<d;k++)
                Ktp += (x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k]);
            Ktp  = -0.5*Ktp*EXPSIG;
            Ktemp =exp(Ktp);

            Kb += 2*Ktemp;
            dKb += 2*Ktemp*Ktp;
            Kib += 2*Ki;
        }
    }
    Kb = Kb/n/n;
    Kib = Kib/n/n;
    dKb = dKb/n/n;

    /* Compute the objective and gradient function */
     #pragma omp parallel for private(i,j,k,Ki,Ktp,logKtp) shared(x,y,Kb,dKb,Kib) reduction(+:sum,sum1,sum2,sum3,sum4) num_threads(MAX_THREADS) schedule(runtime)
    for(i=0;i<n;i++){
        sum += (1-Kib)*(-dKb);
        sum1 += (1-Kb)*(1-Kb);
        sum2 += (1-Kb)*(1-Kib);
        sum3 += (1-Kb)*(-dKb);
        sum4 += (1-Kib)*(1-Kib);
        for(j=i+1;j<n;j++){
            Ki = (y[i] == y[j]) ? 1 : 0;
            Ktp = 0;
            for(k=0;k<d;k++)
                Ktp += (x[i*d+k]-x[j*d+k])*(x[i*d+k]-x[j*d+k]);
            logKtp = -0.5*Ktp*EXPSIG;
            Ktp = exp(-0.5*Ktp*EXPSIG);
            
            sum += 2*(Ki-Kib)*(logKtp*Ktp-dKb);
            sum1 += 2*(Ktp-Kb)*(Ktp-Kb);
            sum2 += 2*(Ki-Kib)*(Ktp-Kb);
            sum3 += 2*(Ktp-Kb)*(logKtp*Ktp-dKb);
            sum4 += 2*(Ki-Kib)*(Ki-Kib);
        }
    }
    J[0] = 1-sum2/sqrt(sum1)/sqrt(sum4);
    dJ[0] = -(sum/sqrt(sum1) - sum2*sum3/sqrt(sum1)/sum1)/sqrt(sum4);
}
    