#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <omp.h>
#include "param.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    /* Some initialization */
    int i, j, k, l;
    double *x= mxGetPr(prhs[0]);
    int nx = mxGetN(prhs[0]);
    int d = mxGetM(prhs[0]);
    double *z= mxGetPr(prhs[1]);
    int nz = mxGetN(prhs[1]);
    double *sig = mxGetPr(prhs[2]);
    
    int ds = (d*(d+1))/2;
    double *SIG = (double*)calloc(ds, sizeof(double));
    for(l=0;l<d;l++){ /* Precompute the exponential for speed up*/
        SIG[IND2VAR(l, l, d)] = exp(sig[IND2VAR(l, l, d)]);
        for(k=l+1;k<d;k++)
            SIG[IND2VAR(l, k, d)] = sig[IND2VAR(l, k, d)];
    }
    
    double *K;
    plhs[0]=mxCreateDoubleMatrix(nx, nz, mxREAL);
    K =  mxGetPr(plhs[0]);
    
    double Ktp,temp;
    #pragma omp parallel for private(i, j, k, l, Ktp, temp) shared(x, z, K, nx, nz) num_threads(MAX_THREADS) schedule(static)
    for(j=0;j<nz;j++){
        for(i=0;i<nx;i++){
            Ktp=0;
            for(l=0;l<d;l++){
                temp = (x[i*d+l]-z[j*d+l])*SIG[IND2VAR(l, l, d)];
                for(k=l+1;k<d;k++){
                    temp += (x[i*d+k]-z[j*d+k])*SIG[IND2VAR(l, k, d)];
                }
                Ktp += (temp*temp);
            }
            K[i+j*nx] = exp(-0.5*Ktp);
        }
        
    }
    free(SIG);
}