% SCRIPT ionosphere %
clear all, close all,  clc
METHOD = 'align'; % Select the similitude measure

% Load data %
load ionosphere.data
% Load the 50 different training/validation set %
load perm_ionosphere.mat

% Build the data
y = ionosphere(:,end);
x = ionosphere(:,1:end-1);
x(:,2) = [];
clear ionosphere
[n d] = size(x);

% Init parameters %
REP = 50; % Should be lower or equal to 50
ct = 1;
split =3;
r_sigma = -50:20;

% Scale the data
x= scale(x);

t = floor(n/split);
while (ct<REP)
    % Split the data
    perm = perm_ionosphere(ct,:);
    xt = x(perm(1:t*(split-1)),:);
    yt = y(perm(1:t*(split-1)));
    xT = x(perm(t*(split-1)+1:end),:);
    yT = y(perm(t*(split-1)+1:end));
    
    % ---------------------- %
    % -- SPHERICAL KERNEL -- %
    % ---------------------- %
    
    % Compute the optimal SIG 
    [sig,fx]=kma(xt',yt,METHOD,'rbf_sphere');
    J_s(ct)=fx(end);
    
    % Compute knn
    yp = knn(xt,yt,xT,sig,'sphere');
    
    % Compute the accuracy
    [oa_s(ct), ~, ~, K_s(ct)]=confusion(yT,yp);
    
    % ---------------------- %
    % - ELLIPSOIDAL KERNEL - %
    % ---------------------- %
    
    % Compute the optimal SIG 
    [sig,fx]=kma(xt',yt,METHOD,'rbf_ellipse');
    J_e(ct)=fx(end);
    
    % Compute knn
    yp = knn(xt,yt,xT,sig,'ellipse');
        
    % Compute the accuracy
    [oa_e(ct), ~, ~, K_e(ct)]=confusion(yT,yp);
    
    % ---------------------- %
    % -- CHOLESKI KERNEL --- %
    % ---------------------- %
    
    % Compute the optimal SIG 
    [sig,fx]=kma(xt',yt,METHOD,'rbf_chol');
    J_c(ct)=fx(end);
    % Compute knn
    yp = knn(xt,yt,xT,sig,'chol');
    
    % Compute the accuracy
    [oa_c(ct), ~, ~, K_c(ct)]=confusion(yT,yp);
    
    % ---------------------- %
    % -------- CV  --------- %
    % ---------------------- %
    [sig cv_m cv]=cross_validation_knn(xt,yt,r_sigma,5);
    yp = knn(xt,yt,xT,sig,'sphere');
    oa_cv(ct)=confusion(yT,yp);
    ct=ct+1;
end
name = sprintf('simu_ionosphere_%s.mat',METHOD);
save(name,'oa_*');